package com.bocloud.cos.booter.scheduler;

import com.bocloud.common.utils.ListTool;
import com.bocloud.cos.common.utils.DateFormatter;
import com.bocloud.cos.common.utils.DateUtils;
import com.bocloud.cos.entity.PublishInfo;
import com.bocloud.cos.entity.ReportConfigTemplate;
import com.bocloud.cos.entity.ReportTemplateHistory;
import com.bocloud.cos.enums.ReportCategory;
import com.bocloud.cos.enums.ReportFileFormat;
import com.bocloud.cos.enums.ReportFrequenceType;
import com.bocloud.cos.intf.*;
import com.bocloud.cos.repository.PublishInfoRepository;
import com.bocloud.cos.repository.ReportConfigTemplateRepository;
import com.bocloud.cos.repository.ReportResourceRepository;
import com.bocloud.cos.repository.ReportTemplateHistoryRepository;
import com.bocloud.service.core.CurrentService;
import com.bocloud.internel.entity.User;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import net.sf.jasperreports.export.ExporterInputItem;
import org.apache.commons.lang.exception.ExceptionUtils;
import org.apache.commons.lang3.StringUtils;
import org.joda.time.DateTime;
import org.joda.time.DateTimeConstants;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.*;

/**
 * @author hufy
 */
@Component
public class ReportScheduler {
    ThreadFactory namedThreadFactory = new ThreadFactoryBuilder().setNameFormat("report-schedule-pool-%d").build();
    ExecutorService sendMailPool = new ThreadPoolExecutor(4, 4, 0L, TimeUnit.MILLISECONDS,
            new LinkedBlockingQueue<>(), namedThreadFactory);

    private Logger logger = LoggerFactory.getLogger(ReportScheduler.class);

    @Autowired
    private CurrentService service;

    @Autowired
    private ReportConfigTemplateRepository reportConfigTemplateDao;

    @Autowired
    private ReportTemplateHistoryRepository reportTemplateHistoryDao;

    @Autowired
    private PublishInfoRepository publishInfoDao;

    @Autowired
    private PlatformInfoReportService platformInfoReportService;
    @Autowired
    private CloudServerReportService cloudServerReportService;
    @Autowired
    private ServerReportService serverReportService;
    @Autowired
    private StorageReportService storageReportService;
    @Autowired
    private ChargeReportService chargeReportService;
    @Autowired
    private MonitorReportService monitorReportService;
    @Autowired
    private TenantReportService tenantReportService;

    @Autowired
    private EmailSender emailSender;

    @Value("${upload.storage.path:/home/cmp}")
    private String directoryPath;

    @Autowired
    private ReportResourceRepository reportResourceDao;

    /**
     * 每30分钟执行一次
     */
    @SuppressWarnings("unused")
    @Scheduled(cron = "${report.history.schedule:0 0/10 * * * ?}")
    protected void sync() {
        Integer pageSize = 50;
        Integer totalPage = null;
        Date current = new Date(System.currentTimeMillis());
        //logger.info("执行报表触发器" + DateFormatter.formatToString(current, "yyyy年MM月dd日_HHmmss"));
        Map<String, Object> params = new HashMap<>(16);
        if (service.isLeader()) {
            // 查询模板
            int count = 0;
            try {
                // 查询总页数
                count = reportConfigTemplateDao.countForSchedule();
            } catch (Exception e) {
                logger.error("查询数量失败", e);
            }
            totalPage = (count + pageSize - 1) / pageSize;

            for (int i = 1; i <= totalPage; i++) {
                try {
                    List<ReportConfigTemplate> templates = reportConfigTemplateDao.listForSchedule(i, pageSize);
                    // 比较触发时间与当前时间，如果到达时间，插入数据，重复的数据，忽略
                    resolveTemplates(templates);
                } catch (Exception e) {
                    logger.error(ExceptionUtils.getFullStackTrace(e));
                }
            }

        }
    }

    /**
     * 处理模板
     *
     * @since 2018-10-19 modify by sxp 增加有效起始日期、有效终止日期的判断
     */
    private void resolveTemplates(List<ReportConfigTemplate> templates) {
        Date today = DateUtils.getToday();

        // 根据模板，计算触发时间
        for (ReportConfigTemplate template : templates) {
            if (!template.getStatus()) {
                continue;
            }
            try {
                // 有效起始日期
                Date beginDate = template.getBeginDate();
                // 有效终止日期
                Date finishDate = template.getFinishDate();
                if (beginDate != null && finishDate != null) {
                    if (today.before(beginDate) || today.after(finishDate)) {
                        continue;
                    }
                }

                // 获取生成时间
                String buildTime = template.getBuildTime();
                String cycleType = template.getCycleType();
                ReportFrequenceType frequenceType = ReportFrequenceType.valueOf(cycleType);
                DateTime now = null;
                String[] dayAndTime = null;
                String dayStr = null;
                String timeStr = null;
                String[] hourAndMi = null;
                String hourStr = null;
                String minuteStr = null;
                int day;
                int hour;
                int minute;

                switch (frequenceType) {
                    // 月报
                    case ByMonth:
                        // 例如 2,12:30 每月2号的12点30
                        dayAndTime = buildTime.split(",");
                        dayStr = dayAndTime[0];
                        timeStr = dayAndTime[1];
                        hourAndMi = timeStr.split(":");
                        hourStr = hourAndMi[0];
                        minuteStr = hourAndMi[1];
                        day = Integer.parseInt(dayStr);
                        hour = Integer.parseInt(hourStr);
                        minute = Integer.parseInt(minuteStr);
                        now = DateTime.now();
                        if (isMonthMatch(now, day, hour, minute, new DateTime(template.getGmtCreate().getTime()))) {
                            // 如果当前时间大于触发时间，先查询改存放的数据是否存在，如果存在则不插入，不存在则插入
                            DateTime triggerTime = DateTime.now().withDayOfMonth(day).withHourOfDay(hour)
                                    .withMinuteOfHour(minute);
                            Date theDate = triggerTime.toDate();
                            saveHistory(template, theDate);
                        }
                        break;
                    // 周报
                    case ByWeek:
                        // 例如 2,12:30 每周2的的12点30
                        dayAndTime = buildTime.split(",");
                        dayStr = dayAndTime[0];
                        timeStr = dayAndTime[1];
                        hourAndMi = timeStr.split(":");
                        hourStr = hourAndMi[0];
                        minuteStr = hourAndMi[1];
                        day = Integer.parseInt(dayStr);
                        hour = Integer.parseInt(hourStr);
                        minute = Integer.parseInt(minuteStr);
                        now = DateTime.now();
                        if (isWeekMatch(now, day, hour, minute, new DateTime(template.getGmtCreate().getTime()))) {
                            // 如果当前时间大于触发时间，先查询改存放的数据是否存在，如果存在则不插入，不存在则插入
                            DateTime triggerTime = DateTime.now().withDayOfWeek(day).withHourOfDay(hour)
                                    .withMinuteOfHour(minute);
                            Date theDate = triggerTime.toDate();
                            saveHistory(template, theDate);
                        }
                        break;
                    // 日报
                    case ByDay:
                        // 例如 12:30 每天的12点30
                        timeStr = buildTime;
                        hourAndMi = timeStr.split(":");
                        hourStr = hourAndMi[0];
                        minuteStr = hourAndMi[1];
                        hour = Integer.parseInt(hourStr);
                        minute = Integer.parseInt(minuteStr);
                        now = DateTime.now();
                        DateTime templateTime = new DateTime(template.getGmtCreate().getTime());
                        if (now.getHourOfDay() >= hour && now.getMinuteOfHour() >= minute) {
                            // 如果当前时间大于触发时间，先查询改存放的数据是否存在，如果存在则不插入，不存在则插入
                            DateTime triggerTime = DateTime.now().withHourOfDay(hour).withMinuteOfHour(minute);
                            Date theDate = triggerTime.toDate();
                            if (triggerTime.isAfter(templateTime)) {
                                saveHistory(template, theDate);
                            }
                        } else {
                            // 看看昨天有没有漏掉的
                            DateTime triggerTime = DateTime.now().plusDays(-1).withHourOfDay(hour).withMinuteOfHour(minute);
                            Date theDate = triggerTime.toDate();

                            // 不能生成比模板创建时间早的 并且生成时间要在模板创建时间的第二天
                            if (triggerTime.isAfter(templateTime)) {
                                saveHistory(template, theDate);
                            }
                        }
                        break;
                    default:
                        break;
                }
            } catch (Exception e) {
                logger.error(ExceptionUtils.getFullStackTrace(e));
                logger.error("出错的模板" + template.getId());
            }
        }

    }

    private boolean isWeekMatch(DateTime now, int day, int hour, int minute, DateTime templateCreateTime) {
        // JodaTime 星期日 对应的是数字 7
        if (day == 0) {
            day = DateTimeConstants.SUNDAY;
        }
        // 因为是次周生成，生成时间所在的周必须在模板创建时间所在周之后
        if (templateCreateTime.getYear() == now.getYear()
                && templateCreateTime.getWeekOfWeekyear() == now.getWeekOfWeekyear()) {
            return false;
        }
        return now.getDayOfWeek() > day
                || (now.getDayOfWeek() == day && now.getHourOfDay() >= hour && now.getMinuteOfHour() >= minute);
    }

    private boolean isMonthMatch(DateTime now, int day, int hour, int minute, DateTime templateCreateTime) {
        // 因为是次月生成，生成时间所在的月必须在模板创建时间所在月之后
        if (templateCreateTime.getYear() == now.getYear()
                && templateCreateTime.getMonthOfYear() == now.getMonthOfYear()) {
            return false;
        }
        return now.getDayOfMonth() > day
                || (now.getDayOfMonth() == day && now.getHourOfDay() >= hour && now.getMinuteOfHour() >= minute);
    }

    /**
     * 生成历史
     *
     * @param template
     * @param theDate
     * @throws Exception
     */
    private void saveHistory(ReportConfigTemplate template, Date theDate) throws Exception {
        Boolean isExist = reportTemplateHistoryDao.queryIfExist(template.getId(), theDate);
        if (!isExist) {
            PublishInfo publishInfo = publishInfoDao.getByConfigId(template.getId());
            String fileFormats = null;
            String userIdStrs = null;
            String[] emails = null;
            if (publishInfo.getStatus()) {
                if (publishInfo != null) {
                    fileFormats = publishInfo.getFileFormat();
                    userIdStrs = publishInfo.getEmail();
                }
                if (userIdStrs == null) {
                    userIdStrs = "";
                }
                emails = userIdsToEmails(userIdStrs.split(","));
            }

            String theDay = DateFormatter.formatToString(theDate, "yyyy年MM月dd日");
            ReportTemplateHistory reportTemplateHistory = new ReportTemplateHistory(template.getId(), theDay, false,
                    theDate, template.getCycleType(), fileFormats, false, null, template.getName() + theDay,
                    template.getCreater());
            reportTemplateHistoryDao.save(reportTemplateHistory);
            Long historyId = reportTemplateHistory.getId();
            // 发送邮件
            if (publishInfo.getStatus()) {
                sendMailPool.submit(new SendMailThread(template, fileFormats, emails, historyId, this));
            } else {
                // 只生成数据
                String reportCat = template.getReportCategory();
                ReportCategory reportCategory = ReportCategory.valueOf(reportCat);
                createReport(historyId, reportCategory);
            }
        }
    }

    private String[] userIdsToEmails(String[] userIdStrs) {
        List<String> strs = Arrays.asList(userIdStrs);
        List<Long> userIds = new ArrayList<>();
        List<String> emails = new ArrayList<>();
        for (String str : strs) {
            if (StringUtils.isNotEmpty(str)) {
                if (StringUtils.isNumeric(str)) {
                    userIds.add(Long.parseLong(str));
                } else {
                    emails.add(str);
                }
            }
        }

        if (emails.size() == 0) {
            return new String[0];
        }

        try {
            if (!ListTool.isEmpty(userIds)) {
                List<User> users = reportResourceDao.queryUsersByUserIds(userIds);
                for (User user : users) {
                    if (StringUtils.isNotBlank(user.getEmail())) {
                        emails.add(user.getEmail());
                    }
                }
            }
        } catch (Exception e) {
            logger.error(ExceptionUtils.getFullStackTrace(e));
        }
        String[] emailArr = new String[emails.size()];
        return emails.toArray(emailArr);
    }

    private String getFileExt(String fileFormat) {
        fileFormat = fileFormat.toUpperCase();
        switch (fileFormat) {
            case "PDF":
                return "pdf";
            case "EXCEL":
                return "xlsx";
            case "WORD":
                return "docx";
            default:
                break;
        }
        return "unknow";
    }

    /**
     * 发送邮件
     *
     * @param template
     * @param fileFormats
     * @param receiverEmails
     * @param historyId
     * @throws IOException
     */
    protected void sendEmail(ReportConfigTemplate template, String fileFormats, String[] receiverEmails, Long historyId)
            throws IOException {
        if (fileFormats != null && receiverEmails != null) {
            String[] allFormat = fileFormats.split(",");
            List<File> files = new ArrayList<>();
            try {

                String reportCat = template.getReportCategory();
                ReportCategory reportCategory = ReportCategory.valueOf(reportCat);
                List<ExporterInputItem> reportList = createReport(historyId, reportCategory);

                for (int i = 0; i < allFormat.length; i++) {
                    String time = new SimpleDateFormat("yyyy年MM月dd日").format(new Date());
                    String fileName = template.getName() + time + "." + getFileExt(allFormat[i]);
                    File file = new File(directoryPath + File.separator + fileName);
                    boolean flag = file.createNewFile();
                    logger.info("Create file result：{}", flag);
                    FileOutputStream outputStream = new FileOutputStream(file);
                    switch (reportCategory) {
                        case Platform:
                            platformInfoReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case CloudServer:
                            cloudServerReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case Server:
                            serverReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case Storage:
                            storageReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case Charge:
                            chargeReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case Monitor:
                            chargeReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        case Tenant:
                            tenantReportService.exportReport(reportList, outputStream,
                                    ReportFileFormat.valueOf(allFormat[i].toUpperCase()));
                            break;
                        default:
                            break;
                    }
                    outputStream.flush();
                    outputStream.close();
                    files.add(file);
                }
                emailSender.send(getName(template.getCycleType()), receiverEmails, "这是您的报表，请查收", files);
            } catch (IllegalArgumentException e) {
                logger.error(ExceptionUtils.getFullStackTrace(e));
            } catch (FileNotFoundException e) {
                logger.error(ExceptionUtils.getFullStackTrace(e));
            } finally {
                for (File file : files) {
                    boolean delete = false;
                    if (file != null) {
                        try {
                            delete = file.delete();
                        } catch (Exception e) {
                            logger.error("删除文件出错", e);
                        }
                        if (!delete) {
                            logger.error("文件没有删除成功" + file.getName());
                        }
                    }
                }

            }
        }
    }

    private List<ExporterInputItem> createReport(Long historyId, ReportCategory reportCategory) {
        List<ExporterInputItem> reportList = null;
        switch (reportCategory) {
            case Platform:
                reportList = platformInfoReportService.createReport(historyId);
                break;
            case CloudServer:
                reportList = cloudServerReportService.createReport(historyId);
                break;
            case Server:
                reportList = serverReportService.createReport(historyId);
                break;
            case Storage:
                reportList = storageReportService.createReport(historyId);
                break;
            case Charge:
                reportList = chargeReportService.createReport(historyId);
                break;
            case Monitor:
                reportList = monitorReportService.createReport(historyId);
                break;
            case Tenant:
                reportList = tenantReportService.createReport(historyId);
                break;
            default:
                break;
        }
        return reportList;
    }

    private String getName(String cycleType) {
        ReportFrequenceType frequenceType = ReportFrequenceType.valueOf(cycleType);
        switch (frequenceType) {
            case ByDay:
                return "日报";
            case ByWeek:
                return "周报";
            case ByMonth:
                return "月报";
            default:
                break;
        }
        return "报表";
    }

    private static class SendMailThread implements Runnable {

        private ReportConfigTemplate template;
        private String fileFormats;
        private String[] emails;
        private Long historyId;
        private ReportScheduler reportScheduler;

        public SendMailThread(ReportConfigTemplate template, String fileFormats, String[] emails, Long historyId,
                              ReportScheduler reportScheduler) {
            this.template = template;
            this.fileFormats = fileFormats;
            this.emails = emails;
            this.historyId = historyId;
            this.reportScheduler = reportScheduler;
        }

        /**
         * When an object implementing interface <code>Runnable</code> is used
         * to create a thread, starting the thread causes the object's
         * <code>run</code> method to be called in that separately executing
         * thread.
         * <p>
         * The general contract of the method <code>run</code> is that it may
         * take any action whatsoever.
         *
         * @see Thread#run()
         */
        @Override
        public void run() {
            try {
                reportScheduler.sendEmail(template, fileFormats, emails, historyId);
            } catch (IOException e) {
                reportScheduler.logger.error(ExceptionUtils.getFullStackTrace(e));
            }
        }
    }
}
