package com.bocloud.cos.repository;

import com.bocloud.common.utils.MapTools;
import com.bocloud.cos.entity.AccountRecord;
import com.bocloud.cos.entity.TenantAccount;
import com.bocloud.database.core.BasicDao;
import com.bocloud.database.core.intf.impl.BasicGenericDao;
import org.springframework.jdbc.core.JdbcTemplate;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.stereotype.Repository;
import org.springframework.util.Assert;
import org.apache.commons.lang3.StringUtils;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * 租户授信账户数据操作类
 *
 * @author tyl
 * @time 2020/7/15 周三 10:04
 */
@Repository("tenantAccountRepository")
public class TenantAccountRepository extends BasicGenericDao<TenantAccount, Long> {
	private final BasicDao basicDao;

	public TenantAccountRepository(JdbcTemplate jdbcTemplate, NamedParameterJdbcTemplate npJdbcTemplate,
			BasicDao basicDao) {
		super(jdbcTemplate, npJdbcTemplate);
		this.basicDao = basicDao;
	}

	/**
	 * 根据租户id删除授信账户
	 *
	 * @param tenantId
	 * @return
	 */
	public boolean delete(Long tenantId) {
		String sql = "delete from tenant_account where tenant_id = :tenantId";
		Map<String, Object> params = MapTools.simpleMap("tenantId", tenantId);
		return execute(sql, params) > 0;
	}

	/**
	 * 根据租户id获取授信账户信息
	 *
	 * @param tenantId
	 * @return
	 */
	public TenantAccount getAccount(Long tenantId) {
		String sql = "select a.*,t.name tenant_name from tenant_account a left join tenant t on a.tenant_id = t.id where a.tenant_id = :tenantId";
		Map<String, Object> params = MapTools.simpleMap("tenantId", tenantId);
		List<TenantAccount> tenantAccounts = this.list(TenantAccount.class, sql, params);
		Assert.isTrue(tenantAccounts.size() <= 1, "more than one tenant account in the table");
		return tenantAccounts.isEmpty() ? null : tenantAccounts.get(0);
	}

	/**
	 * 保存租户授信账户操作记录
	 *
	 * @param accountRecord
	 * @return
	 */
	public AccountRecord saveAccountRecord(AccountRecord accountRecord) {
		return (AccountRecord) this.basicDao.saveEntity(accountRecord);
	}

	/**
	 * 删除租户授信账户操作记录
	 *
	 * @param id
	 * @return
	 */
	public boolean deleteAccount(Long id) {
		String sql = "delete from  account_record where id = :id";
		Map<String, Object> params = MapTools.simpleMap("id", id);
		return this.basicDao.update(sql, params) > 0;
	}

	/**
	 * 根据租户id获取授信账号操作记录
	 *
	 * @param tenantId
	 * @return
	 */
	public List<AccountRecord> getRecords(Long tenantId) {
		String sql = "select a.*,t.name tenant_name from account_record a left join tenant t on a.tenant_id = t.id where a.tenant_id = :tenantId";
		Map<String, Object> params = MapTools.simpleMap("tenantId", tenantId);
		return this.basicDao.queryForList(sql, params, AccountRecord.class).stream()
				.map(object -> (AccountRecord) object).collect(Collectors.toList());
	}

	/**
	 * 根据条件获取授信账号操作记录
	 *
	 * @param tenantId
	 * @return
	 */
	public List<AccountRecord> getRecords(Long tenantId, TenantAccount.Action action, String start, String end) {
		StringBuilder builder = new StringBuilder(
				"select a.*,t.name tenant_name from account_record a left join tenant t on a.tenant_id = t.id where 1=1");
		HashMap<String, Object> params = new HashMap<>();
		if (null != tenantId && 0L != tenantId) {
			builder.append(" and a.tenant_id = :tenantId");
			params.put("tenantId", tenantId);
		}
		if (null != action) {
			builder.append(" and a.action = :action");
			params.put("action", action.name());
		}
		if (StringUtils.isNotEmpty(start)) {
			builder.append(" and a.gmt_create >= :start");
			params.put("start", start);
		}
		if (StringUtils.isNotEmpty(end)) {
			builder.append(" and a.gmt_create <= :end");
			params.put("end", end);
		}
		return this.basicDao.queryForList(builder.toString(), params, AccountRecord.class).stream()
				.map(object -> (AccountRecord) object).collect(Collectors.toList());
	}
}
