package com.bocloud.cos.service;

import com.bocloud.common.model.*;
import com.bocloud.common.utils.ListTool;
import com.bocloud.cos.entity.CloudService;
import com.bocloud.cos.entity.ServiceCatalog;
import com.bocloud.cos.entity.ServiceSla;
import com.bocloud.cos.entity.ServiceTenant;
import com.bocloud.cos.intf.ServiceCatalogService;
import com.bocloud.cos.repository.CloudServiceRepository;
import com.bocloud.cos.repository.OrderSnRepository;
import com.bocloud.cos.repository.ServiceCatalogRepository;
import com.bocloud.cos.repository.ServiceSlaRepository;
import com.google.common.collect.Maps;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * 服务目录业务操作实现
 *
 * @author wangyu
 * @version 1.0
 * @since 2019年5月10日
 */
@Service("serviceCatalogService")
public class ServiceCatalogServiceImpl implements ServiceCatalogService {

    private static final Logger logger = LoggerFactory.getLogger(ServiceCatalogServiceImpl.class);
    private final ServiceCatalogRepository serviceCatalogRepository;
    private final ServiceSlaRepository serviceSlaRepository;
    private final CloudServiceRepository cloudServiceRepository;
    private final OrderSnRepository snRepository;

    public ServiceCatalogServiceImpl(ServiceCatalogRepository serviceCatalogRepository, ServiceSlaRepository serviceSlaRepository, CloudServiceRepository cloudServiceRepository, OrderSnRepository snRepository) {
        this.serviceCatalogRepository = serviceCatalogRepository;
        this.serviceSlaRepository = serviceSlaRepository;
        this.cloudServiceRepository = cloudServiceRepository;
        this.snRepository = snRepository;
    }

    @Override
    public BocloudResult<GridBean<ServiceCatalog>> list(Pager pager) {
        try {
            Map<String, String> sorter = Optional.ofNullable(pager.getSorter()).orElse(Maps.newHashMap());
            List<Param> params = Optional.ofNullable(pager.getParams()).orElse(new ArrayList<>());
            int total = this.serviceCatalogRepository.count(params);
            List<ServiceCatalog> list = this.serviceCatalogRepository.list(pager.getPage(), pager.getRows(), sorter, params);
            GridBean<ServiceCatalog> gridBean = new GridBean<>(pager.getPage(), pager.getRows(), total, list);
            return new BocloudResult<>(true, gridBean, "查询成功");
        } catch (Exception e) {
            logger.error("Query error:", e);
            return new BocloudResult<>(false, "查询失败:" + e.getMessage());
        }
    }

    @Override
    public Result remove(Long id, Long userId) {
        try {
            Boolean result = serviceCatalogRepository.remove(id, userId);
            if (result) {
                snRepository.removeByCatalogId(id);
            }
            return new Result(true, "删除成功");
        } catch (Exception e) {
            logger.error("Query error:", e);
            return new Result(false, "删除失败");
        }
    }

    @Override
    public BocloudResult<ServiceCatalog> detail(Long id) {
        try {
            ServiceCatalog serviceCatalog = serviceCatalogRepository.query(id);
            return new BocloudResult<>(true, serviceCatalog, "查询成功");
        } catch (Exception e) {
            logger.error("Query error:", e);
            return new BocloudResult<>(false, "查询失败:" + e.getMessage());
        }
    }

    @Override
    public Result create(ServiceCatalog serviceCatalog) {
        try {
            Integer maxId = serviceCatalogRepository.getMaxId();
            if (null == maxId) {
                maxId = 0;
            }
            serviceCatalog.setSn(String.format("%02d", maxId));
            serviceCatalogRepository.save(serviceCatalog);
            return new Result(true, "新增成功");
        } catch (Exception e) {
            logger.error("Query error:", e);
            return new Result(false, "新增失败");
        }
    }

    @Override
    public BocloudResult list(Long tenantId) {
        try {
            if (null == tenantId) {
                return new BocloudResult<>(false, "当前租户状态异常，请重新操作！");
            }
            BocloudResult<GridBean<ServiceCatalog>> bocloudResult = this.list(new Pager());
            if (bocloudResult.isFailed()) {
                return bocloudResult;
            }
            GridBean<ServiceCatalog> gridBean = bocloudResult.getData();
            List<ServiceCatalog> serviceCatalogList = gridBean.getRows();
            if (null == serviceCatalogList || serviceCatalogList.size() == 0) {
                return new BocloudResult<>(false, "没有可用服务！");
            }
            for (int i = serviceCatalogList.size() - 1; i >= 0; i--) {
                ServiceCatalog serviceCatalog = serviceCatalogList.get(i);
                List<CloudService> cloudServices = cloudServiceRepository.listByCategory(serviceCatalog.getId(),
                        tenantId);
                if (null != cloudServices && cloudServices.size() > 0) {
                    serviceCatalog.setCloudServices(cloudServices);
                } else {
                    serviceCatalogList.remove(serviceCatalog);
                }
            }
            return new BocloudResult<>(true, serviceCatalogList, "查询成功");
        } catch (Exception e) {
            logger.error("Query error:", e);
            return new BocloudResult<>(false, "查询失败:" + e.getMessage());
        }
    }

    @Override
    public BocloudResult<List<CloudService>> listServiceByCatalogCode(String code, RequestContext context) {
        try {
            ServiceCatalog serviceCatalog = serviceCatalogRepository.getByCode(code);
            List<Long> serviceIds;
            if (context.getCatalog().equals(RequestContext.Catalog.Manager)) {
                serviceIds = cloudServiceRepository.listByStatus(null).stream().map(CloudService::getId).collect(Collectors.toList());
            }else {
                serviceIds = cloudServiceRepository.listByTenantId(context.getTenant()).stream().map(ServiceTenant::getServiceId).collect(Collectors.toList());
            }
            List<CloudService> cloudServices = cloudServiceRepository.listByCategory(serviceCatalog.getId()).stream().filter(cloudService -> serviceIds.contains(cloudService.getId())).collect(Collectors.toList());
            for (CloudService service : cloudServices) {
                List<ServiceSla> slas = serviceSlaRepository.listByService(service.getId(), ServiceSla.Status.enable.name());
                if (!ListTool.isEmpty(slas)) {
                    service.setSlaId(slas.get(0).getId());
                }
            }
            return new BocloudResult<>(true, cloudServices, "查询成功");
        } catch (Exception e) {
            logger.error("list service by catalog code error:", e);
            return new BocloudResult<>(false, "查询失败");
        }
    }
}
