package com.bocloud.cos.service.resource.operate;

import com.alibaba.fastjson.JSONArray;
import com.alibaba.fastjson.JSONObject;
import com.bocloud.common.model.BocloudResult;
import com.bocloud.cos.entity.*;
import com.bocloud.cos.enums.OrderItemStatus;
import com.bocloud.cos.enums.ResourceCategory;
import com.bocloud.cos.intf.OrderSnService;
import com.bocloud.cos.model.*;
import com.bocloud.cos.repository.*;
import com.bocloud.cos.service.utils.IdGenerator;
import com.bocloud.service.lock.BocloudACLock;
import com.bocloud.service.lock.LockFactory;
import com.google.common.collect.Lists;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.List;
import java.util.concurrent.TimeUnit;

/**
 * paas资源操作抽象类
 *
 * @author tyl
 * @since 2021/5/21 9:42
 */
public abstract class AbstractIaasResourceOperate extends AbstractResourceOperate {

    /**
     * 实例id
     */
    String instance;
    /**
     * 资源项
     */
    List<ElementModel> elements;
    /**
     * 延期时长
     */
    Integer extension;
    /**
     * 时长 订购参数
     */
    EmptionModel emptionModel;
    /**
     * 变更操作类型
     */
    AlterationParam.Operation operation;
    /**
     * 服务编码
     */
    final String serviceCode;
    /**
     * 过期时间
     */
    Date expiredTime;
    final Logger logger = LoggerFactory.getLogger(AbstractIaasResourceOperate.class);
    CloudServiceRepository serviceRepository;
    ServiceSkuRepository skuRepository;
    OrderItemRepository itemRepository;
    ServiceOrderRepository orderRepository;
    ServiceCategoryRepository categoryRepository;
    LockFactory lockFactory;
    OrderSnService orderSnService;

    /**
     * 资源申请操作的构造
     *
     * @param serviceCode
     * @param category
     * @param emptionModel
     * @param elements
     * @param serviceRepository
     * @param skuRepository
     */
    public AbstractIaasResourceOperate(String serviceCode, ResourceCategory category, EmptionModel emptionModel,
        List<ElementModel> elements, CloudServiceRepository serviceRepository, ServiceSkuRepository skuRepository,
        ServiceCategoryRepository categoryRepository, LockFactory lockFactory, OrderSnService orderSnService) {
        super(category);
        this.elements = elements;
        this.emptionModel = emptionModel;
        this.serviceCode = serviceCode;
        this.serviceRepository = serviceRepository;
        this.skuRepository = skuRepository;
        this.categoryRepository = categoryRepository;
        this.lockFactory = lockFactory;
        this.orderSnService = orderSnService;
    }

    /**
     * 资源变更操作的构造
     *
     * @param serviceCode
     * @param instance
     * @param category
     * @param operation
     * @param elements
     * @param serviceRepository
     * @param skuRepository
     * @param itemRepository
     */
    public AbstractIaasResourceOperate(String serviceCode, String instance, ResourceCategory category,
        AlterationParam.Operation operation, List<ElementModel> elements, Date expiredTime,
        CloudServiceRepository serviceRepository, ServiceSkuRepository skuRepository,
        OrderItemRepository itemRepository, LockFactory lockFactory, OrderSnService orderSnService,
        ServiceCategoryRepository categoryRepository) {
        super(category);
        this.instance = instance;
        this.elements = elements;
        this.operation = operation;
        this.serviceCode = serviceCode;
        this.expiredTime = expiredTime;
        this.serviceRepository = serviceRepository;
        this.skuRepository = skuRepository;
        this.itemRepository = itemRepository;
        this.lockFactory = lockFactory;
        this.orderSnService = orderSnService;
        this.categoryRepository = categoryRepository;
    }

    /**
     * 资源延期操作的构造
     *
     * @param category
     * @param instance
     * @param extension
     * @param serviceCode
     * @param itemRepository
     * @param orderRepository
     */
    public AbstractIaasResourceOperate(String serviceCode, String instance, Integer extension,
        ResourceCategory category, OrderItemRepository itemRepository, ServiceOrderRepository orderRepository,
        LockFactory lockFactory, OrderSnService orderSnService) {
        super(category);
        this.instance = instance;
        this.extension = extension;
        this.serviceCode = serviceCode;
        this.itemRepository = itemRepository;
        this.orderRepository = orderRepository;
        this.lockFactory = lockFactory;
        this.orderSnService = orderSnService;
    }

    public AbstractIaasResourceOperate(ResourceCategory category, String serviceCode) {
        super(category);
        this.serviceCode = serviceCode;
    }

    @Override public List<OrderItem> buildApplyItem() {
        List<OrderItem> items = new ArrayList<>();
        try {
            DurationModel duration = emptionModel.getDuration();
            Assert.notEmpty(elements, "订购内容不能为空");
            // 订购数量
            Integer count = emptionModel.getCount();
            //计算资源消耗
            Assert.isTrue(null != count && count > 0, "订购数量不能小于0");
            //处理订单项内容和配额
            for (int i = 0; i < count; i++) {
                for (ElementModel element : elements) {
                    String serviceCode = element.getServiceCode();
                    CloudService elementService = serviceRepository.queryByCode(serviceCode);
                    ServiceSku sku = skuRepository.query(element.getSkuId());
                    BigDecimal price = OrderPriceUtils.getUnitPrice(elementService, sku, element.getSpecs(), duration);
                    element.setPrice(price);
                    String elementServiceCode = element.getServiceCode();
                    // 构建订单内容 设置服务类型名称
                    if (null != element.getCategoryId() && 0 != element.getCategoryId()) {
                        ServiceCategory category = categoryRepository.query(element.getCategoryId());
                        element.setCategory(category.getName());
                    } else {
                        element.setCategory(elementService.getName());
                    }
                    OrderItem orderItem = new OrderItem();
                    orderItem.setSkuId(element.getSkuId());
                    String itemSn = this.getItemSn(element.getSkuId());
                    Assert.notNull(itemSn, "生成订购单号失败");
                    orderItem.setSn(itemSn);
                    orderItem.setCatalog(OrderItem.Catalog.single);
                    orderItem.setServiceCode(elementServiceCode);
                    orderItem.setStatus(OrderItemStatus.Opening);
                    orderItem.setDuration(JSONObject.toJSONString(duration));
                    orderItem.setDeleted(false);
                    orderItem.setResources("[]");
                    orderItem.setElements(JSONObject.toJSONString(Lists.newArrayList(element)));
                    orderItem.setBasicPrice(price);
                    items.add(orderItem);
                }
            }
        } catch (Exception e) {
            logger.error("build apply item error", e);
        }
        return items;
    }

    @Override public List<OrderItem> buildAlterationItem() {
        try {
            Assert.notEmpty(elements, "变更内容不能为空");
            ElementModel elementModel = elements.get(0);
            Calendar expiredCalendar = Calendar.getInstance();
            Calendar effectiveCalendar = Calendar.getInstance();
            switch (operation) {
                case AddDisk:
                    OrderItem serverRelationItem = itemRepository.queryLastSuccessBySn(instance);
                    DurationModel addDuration =
                        JSONObject.parseObject(serverRelationItem.getDuration(), DurationModel.class);
                    //预付费
                    if (addDuration.getAmount()!=null && addDuration.getAmount() > 0) {
                        expiredCalendar.setTime(expiredTime);
                        effectiveCalendar.setTime(new Date());
                        if (addDuration.getMode().equals(DurationModel.Mode.Month)) {
                            int expiredYear = expiredCalendar.get(Calendar.YEAR);
                            int effectiveYear = effectiveCalendar.get(Calendar.YEAR);
                            int expired = expiredCalendar.get(Calendar.MONTH);
                            int effective = effectiveCalendar.get(Calendar.MONTH);
                            addDuration.setAmount((expiredYear - effectiveYear) * 12 + (expired - effective));
                        } else if (addDuration.getMode().equals(DurationModel.Mode.Year)) {
                            int expired = expiredCalendar.get(Calendar.YEAR);
                            int effective = effectiveCalendar.get(Calendar.YEAR);
                            addDuration.setAmount(expired - effective);
                        }
                    }
                    return this.handlerAddDiskItems(addDuration, elements);
                case EditDisk:
                    String uuid = elementModel.getUuid();
                    OrderItem diskRelationItem = itemRepository.queryLastSuccessBySn(uuid);
                    DurationModel editDuration =
                        JSONObject.parseObject(diskRelationItem.getDuration(), DurationModel.class);
                    if (editDuration.getAmount()!=null && editDuration.getAmount() > 0) {
                        expiredCalendar.setTime(expiredTime);
                        effectiveCalendar.setTime(new Date());
                        if (editDuration.getMode().equals(DurationModel.Mode.Month)) {
                            int expiredYear = expiredCalendar.get(Calendar.YEAR);
                            int effectiveYear = effectiveCalendar.get(Calendar.YEAR);
                            int expired = expiredCalendar.get(Calendar.MONTH);
                            int effective = effectiveCalendar.get(Calendar.MONTH);
                            editDuration.setAmount((expiredYear - effectiveYear) * 12 + (expired - effective));
                        } else if (editDuration.getMode().equals(DurationModel.Mode.Year)) {
                            int expired = expiredCalendar.get(Calendar.YEAR);
                            int effective = effectiveCalendar.get(Calendar.YEAR);
                            editDuration.setAmount(expired - effective);
                        }
                    }
                    OrderItem editItem = this.handlerEditItem(editDuration, uuid, elementModel);
                    return Lists.newArrayList(editItem);
                case Upgrade:
                    OrderItem relationItem = itemRepository.queryLastSuccessBySn(instance);
                    DurationModel upDuration = JSONObject.parseObject(relationItem.getDuration(), DurationModel.class);
                    if (upDuration.getAmount()!=null && upDuration.getAmount() > 0) {
                        expiredCalendar.setTime(expiredTime);
                        effectiveCalendar.setTime(new Date());
                        if (upDuration.getMode().equals(DurationModel.Mode.Month)) {
                            int expiredYear = expiredCalendar.get(Calendar.YEAR);
                            int effectiveYear = effectiveCalendar.get(Calendar.YEAR);
                            int expired = expiredCalendar.get(Calendar.MONTH);
                            int effective = effectiveCalendar.get(Calendar.MONTH);
                            upDuration.setAmount((expiredYear - effectiveYear) * 12 + (expired - effective));
                        } else if (upDuration.getMode().equals(DurationModel.Mode.Year)) {
                            int expired = expiredCalendar.get(Calendar.YEAR);
                            int effective = effectiveCalendar.get(Calendar.YEAR);
                            upDuration.setAmount(expired - effective);
                        }
                    }
                    OrderItem upgradeItem = handlerUpgradeItem(upDuration, instance, elementModel);
                    return Lists.newArrayList(upgradeItem);
                default:
                    break;
            }
            return null;
        } catch (Exception e) {
            logger.error("build alteration item error,", e);
            return null;
        }
    }

    private List<OrderItem> handlerAddDiskItems(DurationModel duration, List<ElementModel> elements) {
        List<OrderItem> items = new ArrayList<>();
        for (ElementModel element : elements) {
            element.setCategory(element.getCategory());
            //变更的服务
            Long alterationSkuId = element.getSkuId();
            String serviceCode = element.getServiceCode();
            CloudService elementService = serviceRepository.queryByCode(serviceCode);
            ServiceSku sku = skuRepository.query(element.getSkuId());
            BigDecimal price = OrderPriceUtils.getUnitPrice(elementService, sku, element.getSpecs(), duration);
            element.setPrice(price);
            //构建新添加磁盘的订单项
            OrderItem addItem = new OrderItem();
            addItem.setSkuId(alterationSkuId);
            String addSn = this.getItemSn(alterationSkuId);
            Assert.notNull(addSn, "生成订购单号失败");
            addItem.setSn(addSn);
            element.setUuid(addSn);
            addItem.setServiceCode(element.getServiceCode());
            addItem.setStatus(OrderItemStatus.Changing);
            addItem.setDuration(JSONObject.toJSONString(duration));
            addItem.setDeleted(false);
            addItem.setResources("[]");
            addItem.setElements(JSONObject.toJSONString(Lists.newArrayList(element)));
            addItem.setBasicPrice(price);
            addItem.setEffectiveTime(new Date());
            addItem.setExpiredTime(expiredTime);
            items.add(addItem);
        }
        return items;
    }

    private OrderItem handlerEditItem(DurationModel duration, String uuid, ElementModel element) {
        element.setCategory(element.getCategory());
        //变更的服务
        Long alterationSkuId = element.getSkuId();
        //获取价格
        String serviceCode = element.getServiceCode();
        CloudService elementService = serviceRepository.queryByCode(serviceCode);
        ServiceSku sku = skuRepository.query(alterationSkuId);
        BigDecimal price = OrderPriceUtils.getUnitPrice(elementService, sku, element.getSpecs(), duration);
        element.setPrice(price);
        //构建新添加磁盘的订单项
        OrderItem orderItem = new OrderItem();
        orderItem.setSkuId(alterationSkuId);
        orderItem.setSn(uuid);
        orderItem.setServiceCode(element.getServiceCode());
        orderItem.setStatus(OrderItemStatus.Changing);
        orderItem.setDuration(JSONObject.toJSONString(duration));
        orderItem.setDeleted(false);
        orderItem.setResources("[]");
        orderItem.setElements(JSONObject.toJSONString(Lists.newArrayList(element)));
        orderItem.setBasicPrice(price);
        orderItem.setEffectiveTime(new Date());
        orderItem.setExpiredTime(expiredTime);
        return orderItem;
    }

    private OrderItem handlerUpgradeItem(DurationModel duration, String uuid, ElementModel element) {
        element.setCategory(element.getCategory());
        //变更的服务
        Long alterationSkuId = element.getSkuId();
        ServiceSku sku = skuRepository.query(alterationSkuId);
        // 处理规格计算
        List<SpecModel> specModels = element.getSpecs();
        Integer cpuTotal;
        int memTotal;
        if (specModels == null || specModels.isEmpty()) {
            List<SkuSpec> skuSpecs = JSONArray.parseArray(sku.getSpec(), SkuSpec.class);
            cpuTotal = skuSpecs.stream().filter(skuSpec -> "cpu".equals(skuSpec.getSpecName()))
                .map(skuSpec -> Integer.parseInt(skuSpec.getSpecValue())).reduce(Integer::sum).orElse(0);
            memTotal = skuSpecs.stream().filter(skuSpec -> "memory".equals(skuSpec.getSpecName()))
                .map(skuSpec -> Integer.parseInt(skuSpec.getSpecValue())).reduce(Integer::sum).orElse(0);
            List<SpecModel> elementSpec = new ArrayList<>();
            SpecModel cpuSpec = new SpecModel();
            cpuSpec.setCpu(cpuTotal);
            elementSpec.add(cpuSpec);
            SpecModel memSpec = new SpecModel();
            memSpec.setMemory(BigDecimal.valueOf(memTotal));
            elementSpec.add(memSpec);
            element.setSpecs(elementSpec);
        }
        String serviceCode = element.getServiceCode();
        CloudService elementService = serviceRepository.queryByCode(serviceCode);
        BigDecimal price = OrderPriceUtils.getUnitPrice(elementService, sku, element.getSpecs(), duration);
        element.setPrice(price);
        //构建变更规格的订单项
        OrderItem upgradeItem = new OrderItem();
        upgradeItem.setSkuId(alterationSkuId);
        upgradeItem.setSn(uuid);
        upgradeItem.setServiceCode(element.getServiceCode());
        upgradeItem.setStatus(OrderItemStatus.Changing);
        upgradeItem.setDuration(JSONObject.toJSONString(duration));
        upgradeItem.setDeleted(false);
        upgradeItem.setResources("[]");
        upgradeItem.setElements(JSONObject.toJSONString(Lists.newArrayList(elements)));
        upgradeItem.setBasicPrice(price);
        upgradeItem.setEffectiveTime(new Date());
        upgradeItem.setExpiredTime(expiredTime);
        return upgradeItem;
    }

    public String getItemSn(Long skuId) {
        if (null == skuId || 0L == skuId) {
            // 生成订单号，增加分布式锁
            String itemPath = IdGenerator.class.getSimpleName() + "_order_item_sn";
            try (BocloudACLock lock = lockFactory.getACLock(itemPath)) {
                Assert.isTrue(lock.acquire(1, TimeUnit.MINUTES), "生成订单号请求超时");
                return IdGenerator.getInstance().nextId();
            } catch (Exception e) {
                logger.error("Get error message：{}", e.getMessage());
                return null;
            }
        } else {
            BocloudResult<String> snResult = orderSnService.generate(skuId);
            Assert.isTrue(snResult.isSuccess(), snResult.getMessage());
            return snResult.getData();
        }
    }
}
